from vllm import LLM, SamplingParams
from vllm.sampling_params import GuidedDecodingParams
from vllm.outputs import RequestOutput

import torch

import numpy as np

import asyncio
import threading

from openai import AsyncOpenAI
from openai.types.chat import ChatCompletion

from typing import Any, List, Optional, Union, Dict

from .dynamic_pydantic import _generate_pydantic_model
from .response_generation import (
    ResponseGenerationMethod,
    JSONResponseGenerationMethod,
    ChoiceResponseGenerationMethod,
    LogprobResponseGenerationMethod,
)

import json
import random

from tqdm.asyncio import tqdm_asyncio

# @dataclass
# class StructuredOutputOptions:
#     """
#     Configuration for structured output generation.

#     Attributes:
#         category: Type of structured output ("choice" or "json")
#         json_fields: List of field names for JSON output
#         constraints: Optional constraints for field values
#         allowed_choices: List of allowed choices for choice output
#         automatic_system_prompt: If a instruction to only output in the required json format should be added to the system prompt
#     """

#     category: Literal["choice", "json"]
#     json_fields: Optional[List[str]] = None
#     constraints: Optional[Dict[str, List[str]]] = None
#     allowed_choices: Optional[List[str]] = None
#     automatic_system_prompt: bool = False

#     def __post_init__(self):
#         """Perform validation after the object has been initialized."""
#         if self.category == "json" and self.json_fields is None:
#             raise ValueError(
#                 "`json_fields` must be provided when category is 'json'"
#             )

#         if self.category == "choice" and self.allowed_choices is None:
#             raise ValueError(
#                 "`allowed_choices` must be provided when category is 'choice'"
#             )


import regex as re

from tqdm.auto import tqdm


def default_model_init(model_id: str, seed: int = 42, **model_keywords) -> LLM:
    """
    Initialize a vLLM model with default settings.

    Args:
        model_id: HuggingFace model identifier
        seed: Random seed for reproducibility
        **model_keywords: Additional keywords passed to LLM constructor

    Returns:
        LLM: Initialized vLLM model instance
    """
    random.seed(seed)
    torch.manual_seed(seed)
    print("Device_count: " + str(torch.cuda.device_count()))
    print(model_keywords)

    return LLM(
        model=model_id,
        tensor_parallel_size=torch.cuda.device_count(),
        seed=seed,
        **model_keywords,
    )


def _generate_seeds(seed: int, batch_size: int) -> List[int]:
    """
    Generate a list of random seeds.

    Args:
        seed: Base random seed
        batch_size: Number of seeds to generate

    Returns:
        List[int]: Generated random seeds
    """
    rng = np.random.default_rng(seed)
    return rng.integers(low=0, high=2**32, size=batch_size).tolist()


# TODO Structured output for API calls
def batch_generation(
    model: Union[LLM, AsyncOpenAI],
    system_messages: List[str] = ["You are a helpful assistant."],
    prompts: List[str] = ["Hi there! What is your name?"],
    response_generation_method: Optional[
        Union[ResponseGenerationMethod, List[ResponseGenerationMethod]]
    ] = None,
    seed: int = 42,
    client_model_name: Optional[str] = None,
    api_concurrency: int = 10,
    print_conversation: bool = False,
    number_of_printed_conversation: int = 2,
    print_progress: bool = True,
    # <think>...</think> tokens are used by Qwen3 to separate reasoning
    reasoning_start_token: str = "<think>",
    reasoning_end_token: str = "</think>",
    space_char: str = "Ġ",
    chat_template: Optional[str] = None,
    chat_template_kwargs: Dict[str, Any] = {},
    **generation_kwargs: Any,
):
    """
    Generate responses for a batch of prompts.

    Handles both vLLM and OpenAI API generation with support for:
    - Structured output (JSON or choice format)
    - Conversation printing
    - Progress tracking
    - Concurrent API requests

    Args:
        model: vLLM model or AsyncOpenAI client
        system_messages: System prompts for each conversation
        prompts: User prompts to generate responses for
        answer_production_method: Configuration for structured output
        seed: Random seed for reproducibility
        client_model_name: Model name when using OpenAI API
        api_concurrency: Max concurrent API requests
        print_conversation: If True, prints conversations
        print_progress: If True, shows progress bar
        reasoning_start_token: Special token at the beginning of reasoning models' output
        reasoning_end_token: Special token to separate reasoning from regular model output
        space_token: Special char to encode spaces in tokens ("Ġ" for most byte-pair tokenizers)
        **generation_kwargs: Additional generation parameters

    Returns:
        List[str]: Generated responses
    """
    random.seed(seed)

    # Prepare batch of messages
    batch_messages: List[List[Dict[str, str]]] = [
        [
            {"role": "system", "content": system_message},
            {"role": "user", "content": prompt},
        ]
        for system_message, prompt in zip(system_messages, prompts)
    ]

    batch_size: int = len(system_messages)

    seeds = _generate_seeds(seed, batch_size=batch_size)

    logprob_result = None

    if isinstance(model, LLM):
        # TODO: add support for List[AnswerProductionMethod]
        if isinstance(response_generation_method, LogprobResponseGenerationMethod):
            generation_kwargs["logprobs"] = response_generation_method.top_logprobs
            # set token limit separately to support reasoning models
            if response_generation_method.token_limit is not None:
                generation_kwargs["max_tokens"] = response_generation_method.token_limit

        sampling_params_list = _create_sampling_params(
            batch_size=batch_size,
            seeds=seeds,
            response_generation_method=response_generation_method,
            **generation_kwargs,
        )
        outputs: List[RequestOutput] = model.chat(
            batch_messages,
            sampling_params=sampling_params_list,
            use_tqdm=print_progress,
            chat_template=chat_template,
            chat_template_kwargs=chat_template_kwargs,
        )
        result = [output.outputs[0].text for output in outputs]

        # separate out reasoning
        # we parse this here because the OpenAI API separates it automatically
        plain_results = []
        reasoning_output = []
        raw_reasonings = []  # keep the whitespace for length calculations
        for output_text in result:
            reasoning_match = re.search(
                reasoning_start_token + r"(.*?)" + reasoning_end_token,
                output_text,
                re.DOTALL,
            )
            raw_reasonings.append(reasoning_match.group(1) if reasoning_match else None)
            reasoning_output.append(
                reasoning_match.group(1).strip() if reasoning_match else None
            )
            plain_results.append(output_text.split(reasoning_end_token)[-1].strip())

        for rgm in response_generation_method:
            # TODO This is not implemented correcty yet
            if isinstance(rgm, LogprobResponseGenerationMethod):
                logprob_result = []
                # ignore the first k tokens that belong to the reasoning
                if rgm.ignore_reasoning:
                    tokenizer = model.get_tokenizer()
                    logprob_positions = [
                        (
                            len(
                                tokenizer.tokenize(
                                    f"{reasoning_start_token}{_reasoning}{reasoning_end_token}"
                                )
                            )
                            + 1
                            + rgm.token_position
                            if _reasoning is not None
                            else rgm.token_position
                        )
                        for _reasoning in raw_reasonings
                    ]
                else:
                    logprob_positions = [rgm.token_position] * len(outputs)

                for req_output, logprob_position in zip(outputs, logprob_positions):
                    try:
                        answer_dict = {
                            x.decoded_token.lstrip(
                                space_char
                            ).lstrip(): x.logprob  # strip the space character and whitespace from tokenization
                            for x in req_output.outputs[0]
                            .logprobs[logprob_position]
                            .values()
                        }
                    except (
                        IndexError
                    ):  # less than [logprob_position] tokens in the output!
                        answer_dict = {}
                    logprob_result.append(answer_dict)

        # print the first result returned from vllm
        # if print_conversation:
        #    for req_output in outputs:
        #        print(req_output.outputs[0])
        #        break

    else:
        # TODO: add support for List[AnswerProductionMethod]
        if isinstance(response_generation_method[0], LogprobResponseGenerationMethod):
            raise NotImplementedError(
                "The Logprob_AnswerProductionMethod is not yet implemented "
                + "for use with the OpenAI API. Use vllm offline inference instead."
            )
            generation_kwargs["logprobs"] = True
            generation_kwargs["top_logprobs"] = response_generation_method.top_logprobs

        plain_results = _run_async_in_thread(
            client=model,
            client_model_name=client_model_name,
            batch_messages=batch_messages,
            seeds=seeds,
            concurrency_limit=api_concurrency,
            response_generation_method=response_generation_method,
            **generation_kwargs,
        )

        # TODO: handle reasoning
        reasoning_output = [None] * len(plain_results)

    if logprob_result is None:
        logprob_result = [None] * len(plain_results)

    # TODO add argument to specify how many conversations should be printed (base argument should be reasonable)
    if print_conversation:
        conversation_print = "--- Conversation ---"
        for system_message, prompt, answer, reasoning, logprob_answer in zip(
            system_messages, prompts, plain_results, reasoning_output, logprob_result
        ):
            round_print = f"{conversation_print}\n-- System Message --\n{system_message}\n-- User Message ---\n{prompt}\n-- Generated Message --\n{answer}"
            if reasoning is not None:
                round_print += "\n-- Reasoning --\n" + str(reasoning)
            if isinstance(response_generation_method, LogprobResponseGenerationMethod):
                round_print += "\n-- Logprobs --\n" + str(logprob_answer)
            tqdm.write(round_print)

    return (plain_results, logprob_result, reasoning_output)


def _make_cache_key(fields: Any, constraints: Any) -> str:
    return json.dumps({"fields": fields, "constraints": constraints}, sort_keys=False)


def _create_sampling_params(
    batch_size: int,
    seeds: List[int],
    response_generation_method: Optional[
        Union[ResponseGenerationMethod, List[ResponseGenerationMethod]]
    ],
    use_vllm: bool = True,
    **generation_kwargs: Any,
) -> Union[List[SamplingParams], Dict[str, Any]]:
    """
    Create sampling parameters for generation.

    Args:
        batch_size: Number of prompts in batch
        seeds: Random seeds for generation
        answer_production_method: Output structure configuration
        use_vllm: If True, creates vLLM parameters
        **generation_kwargs: Additional sampling parameters

    Returns:
        Sampling parameters for vLLM or API configuration
    """

    if response_generation_method:
        sampling_params_list = _structured_sampling_params(
            batch_size=batch_size,
            seeds=seeds,
            response_generation_method=response_generation_method,
            use_vllm=use_vllm,
            **generation_kwargs,
        )
    else:
        if use_vllm:
            sampling_params_list = [
                SamplingParams(seed=seeds[i], **generation_kwargs)
                for i in range(batch_size)
            ]
        else:
            return []
    return sampling_params_list


def _structured_sampling_params(
    batch_size: int,
    seeds: List[int],
    response_generation_method: Union[
        ResponseGenerationMethod, List[ResponseGenerationMethod]
    ],
    use_vllm: bool = True,
    **generation_kwargs: Any,
) -> Union[List[SamplingParams], Dict[str, Any]]:

    guided_decodings = []

    if isinstance(response_generation_method, ResponseGenerationMethod):
        if isinstance(response_generation_method, JSONResponseGenerationMethod):
            pydantic_model = _generate_pydantic_model(
                fields=response_generation_method.json_fields,
                constraints=response_generation_method.constraints,
            )
            json_schema = pydantic_model.model_json_schema()
            if use_vllm:
                global_guided_decoding = GuidedDecodingParams(json=json_schema)
                guided_decodings = [global_guided_decoding] * batch_size
            else:
                guided_decodings = [json_schema] * batch_size
        elif (
            isinstance(
                response_generation_method,
                (ChoiceResponseGenerationMethod, LogprobResponseGenerationMethod),
            )
            and response_generation_method.allowed_choices is not None
        ):
            _allowed_choices = [
                str(c) for c in response_generation_method.allowed_choices
            ]
            if use_vllm:
                global_guided_decoding = GuidedDecodingParams(choice=_allowed_choices)
                guided_decodings = [global_guided_decoding] * batch_size
            else:
                guided_decodings = [_allowed_choices] * batch_size

    else:
        guided_decodings = []
        cache: Dict[str, GuidedDecodingParams] = {}
        for i in range(batch_size):
            if isinstance(response_generation_method[i], JSONResponseGenerationMethod):
                fields = response_generation_method[i].json_fields
                cons = response_generation_method[i].constraints

                key = _make_cache_key(fields, cons)

                if key not in cache:
                    pydantic_model = _generate_pydantic_model(
                        fields=fields, constraints=cons
                    )
                    json_schema = pydantic_model.model_json_schema()
                    if use_vllm:
                        cache[key] = GuidedDecodingParams(json=json_schema)
                    else:
                        cache[key] = json_schema

                guided_decodings.append(cache[key])
            elif (
                isinstance(
                    response_generation_method[i],
                    (ChoiceResponseGenerationMethod, LogprobResponseGenerationMethod),
                )
                and response_generation_method[i].allowed_choices is not None
            ):
                _allowed_choices = [
                    str(c) for c in response_generation_method[i].allowed_choices
                ]

                key = _make_cache_key(_allowed_choices, None)
                if key not in cache:
                    if use_vllm:
                        cache[key] = GuidedDecodingParams(choice=_allowed_choices)
                    else:
                        cache[key] = _allowed_choices
                guided_decodings.append(cache[key])
            else:
                guided_decodings.append(None)

    if use_vllm and len(guided_decodings) == batch_size:
        sampling_params_list = [
            SamplingParams(
                seed=seeds[i],
                guided_decoding=guided_decodings[i],
                **generation_kwargs,
            )
            for i in range(batch_size)
        ]
    elif use_vllm:
        sampling_params_list = [
            SamplingParams(seed=seeds[i], **generation_kwargs)
            for i in range(batch_size)
        ]
    else:
        return guided_decodings

    return sampling_params_list


def batch_turn_by_turn_generation(
    model: LLM,
    system_messages: List[str] = ["You are a helpful assistant."],
    prompts: List[List[str]] = [["Hi there! What is your name?", "Interesting"]],
    assistant_messages: List[List[str]] = None,
    response_generation_method: Optional[
        Union[ResponseGenerationMethod, List[ResponseGenerationMethod]]
    ] = None,
    seed: int = 42,
    client_model_name: Optional[str] = None,
    api_concurrency: int = 10,
    print_conversation: bool = False,
    print_progress: bool = True,
    reasoning_start_token: str = "<think>",
    reasoning_end_token: str = "</think>",
    space_char: str = "Ġ",
    chat_template: Optional[str] = None,
    chat_template_kwargs: Dict[str, Any] = {},
    **generation_kwargs,
) -> List[str]:
    """
    Generate responses for multi-turn conversations.

    Handles conversations with multiple back-and-forth exchanges between
    user and assistant. Supports:
    - Structured output formats
    - Pre-filled assistant messages
    - Conversation printing
    - Progress tracking

    Args:
        model: vLLM model or AsyncOpenAI client
        system_messages: System prompts for each conversation
        prompts: Lists of user messages for each conversation
        assistant_messages: Optional pre-filled assistant responses
        answer_production_method: Output structure configuration
        seed: Random seed for reproducibility
        client_model_name: Model name for OpenAI API
        api_concurrency: Max concurrent API requests
        print_conversation: If True, prints conversations
        print_progress: If True, shows progress bar
        **generation_kwargs: Additional generation parameters

    Returns:
        List[str]: Generated responses for each conversation
    """

    random.seed(seed)
    batch_messages = []
    batch_size = len(system_messages)
    for i in range(batch_size):
        messages = []

        # Add system message
        if system_messages[i]:
            messages.append({"role": "system", "content": system_messages[i]})

        num_user_msgs = len(prompts[i])
        num_assistant_msgs = len(assistant_messages[i])

        # TODO this implementation is wrong, because assistant messages supports a dict, so they can be anywhere and not just at the beginning
        for j in range(num_user_msgs):
            messages.append({"role": "user", "content": prompts[i][j]})
            if j < num_assistant_msgs:
                messages.append(
                    {"role": "assistant", "content": assistant_messages[i][j]}
                )

        batch_messages.append(messages)

    seeds = _generate_seeds(seed, batch_size=batch_size)

    if isinstance(model, LLM):
        sampling_params_list = _create_sampling_params(
            batch_size=batch_size,
            seeds=seeds,
            response_generation_method=response_generation_method,
            **generation_kwargs,
        )
        outputs: List[RequestOutput] = model.chat(
            batch_messages,
            sampling_params=sampling_params_list,
            use_tqdm=print_progress,
            chat_template=chat_template,
            chat_template_kwargs=chat_template_kwargs,
        )
        # TODO: add support for logprobs
        result = [output.outputs[0].text for output in outputs]

        # separate out reasoning
        # we parse this here because the OpenAI API separates it automatically
        plain_results = []
        reasoning_output = []
        raw_reasonings = []  # keep the whitespace for length calculations
        logprob_result = []
        for output_text in result:
            reasoning_match = re.search(
                reasoning_start_token + r"(.*?)" + reasoning_end_token,
                output_text,
                re.DOTALL,
            )
            raw_reasonings.append(reasoning_match.group(1) if reasoning_match else None)
            reasoning_output.append(
                reasoning_match.group(1).strip() if reasoning_match else None
            )
            plain_results.append(output_text.split(reasoning_end_token)[-1].strip())

        if isinstance(response_generation_method, LogprobResponseGenerationMethod):
            # ignore the first k tokens that belong to the reasoning
            if response_generation_method.ignore_reasoning:
                tokenizer = model.get_tokenizer()
                logprob_positions = [
                    (
                        len(
                            tokenizer.tokenize(
                                f"{reasoning_start_token}{_reasoning}{reasoning_end_token}"
                            )
                        )
                        + 1
                        + response_generation_method.token_position
                        if _reasoning is not None
                        else response_generation_method.token_position
                    )
                    for _reasoning in raw_reasonings
                ]
            else:
                logprob_positions = [response_generation_method.token_position] * len(
                    outputs
                )

            for req_output, logprob_position in zip(outputs, logprob_positions):
                try:
                    answer_dict = {
                        x.decoded_token.lstrip(
                            space_char
                        ).lstrip(): x.logprob  # strip the space character and whitespace from tokenization
                        for x in req_output.outputs[0]
                        .logprobs[logprob_position]
                        .values()
                    }
                except IndexError:  # less than [logprob_position] tokens in the output!
                    answer_dict = {}
                logprob_result.append(answer_dict)

    else:
        plain_results = _run_async_in_thread(
            client=model,
            client_model_name=client_model_name,
            batch_messages=batch_messages,
            seeds=seeds,
            concurrency_limit=api_concurrency,
            response_generation_method=response_generation_method,
            print_progress=print_progress,
            **generation_kwargs,
        )

        # TODO: handle reasoning and logprobs
        reasoning_output = [None] * len(plain_results)
        logprob_result = [None] * len(plain_results)

    # TODO add argument to specify how many conversations should be printed
    if print_conversation:
        conversation_print = "Conversation:"
        for system_message, prompt_list, assistant_list, answer in zip(
            system_messages, prompts, assistant_messages, result
        ):
            round_print = f"{conversation_print}\nSystem Prompt:\n{system_message}"
            for j in range(len(prompt_list)):
                round_print = f"{round_print}\nUser Message:\n{prompt_list[j]}"
                if j < len(assistant_list):
                    prefill = assistant_list[j]
                    if prefill:
                        round_print = (
                            f"{round_print}\nAssistant Message:\n{assistant_list[j]}"
                        )
            round_print = f"{round_print}\nGenerated Answer:\n{answer}"
            tqdm.write(round_print)

    return (plain_results, logprob_result, reasoning_output)


def _run_async_in_thread(
    client: AsyncOpenAI,
    client_model_name: str,
    batch_messages: List[List[Dict[str, str]]],
    seeds: List[int],
    concurrency_limit: int = 10,
    print_progress: bool = True,
    response_generation_method: Optional[
        Union[ResponseGenerationMethod, List[ResponseGenerationMethod]]
    ] = None,
    **generation_kwargs,
):
    result_container = {}

    guided_decoding_params = _create_sampling_params(
        batch_size=len(batch_messages),
        seeds=seeds,
        response_generation_method=response_generation_method,
        use_vllm=False,
        **generation_kwargs,
    )

    def thread_target():
        try:
            res = asyncio.run(
                _run_api_batch_async(
                    client=client,
                    client_model_name=client_model_name,
                    batch_messages=batch_messages,
                    seeds=seeds,
                    concurrency_limit=concurrency_limit,
                    print_progress=print_progress,
                    response_generation_method=response_generation_method,
                    guided_decoding_params=guided_decoding_params,
                    **generation_kwargs,
                )
            )
            result_container["result"] = res
        except Exception as e:
            result_container["error"] = e

    thread = threading.Thread(target=thread_target)
    thread.start()
    thread.join()

    if "error" in result_container:
        raise result_container["error"]

    return result_container.get("result")


async def _run_api_batch_async(
    client: AsyncOpenAI,
    client_model_name: str,
    batch_messages: List[List[Dict[str, str]]],
    seeds: List[int],
    concurrency_limit: int = 10,
    print_progress: bool = True,
    guided_decoding_params: List[Dict[str, Any]] = [],
    response_generation_method: Optional[
        Union[ResponseGenerationMethod, List[ResponseGenerationMethod]]
    ] = None,
    **generation_kwargs,
) -> List[str]:
    semaphore = asyncio.Semaphore(concurrency_limit)

    async def get_completion(
        messages: list,
        seed: int,
        guided_decoding_params: Optional[Union[Dict[str, Any], List[str]]] = None,
        response_generation_method: Optional[
            Union[ResponseGenerationMethod, List[ResponseGenerationMethod]]
        ] = None,
        **generation_kwargs,
    ) -> ChatCompletion:
        async with semaphore:
            request_kwargs = {
                "model": client_model_name,
                "messages": messages,
                "seed": seed,
                **generation_kwargs,
            }

            if response_generation_method:
                if isinstance(response_generation_method, JSONResponseGenerationMethod):
                    request_kwargs["response_format"] = {
                        "type": "json_schema",
                        "json_schema": {
                            "name": "json_schema",
                            "schema": guided_decoding_params,
                        },
                    }
                elif isinstance(
                    response_generation_method, ChoiceResponseGenerationMethod
                ):
                    # TODO: add warning if this is not running against vllm, i.e., guided_choice is not supported
                    request_kwargs["extra_body"] = {
                        "guided_choice": guided_decoding_params
                    }

            return await client.chat.completions.create(**request_kwargs)

    # pbar = tqdm.tqdm if print_progress else lambda x: x

    if len(guided_decoding_params) > 0:
        tasks = [
            get_completion(messages, seed, struct_output, rgm, **generation_kwargs)
            for messages, seed, struct_output, rgm in zip(
                batch_messages,
                seeds,
                guided_decoding_params,
                response_generation_method,
            )
        ]
    else:
        tasks = [
            get_completion(messages, seed, **generation_kwargs)
            for messages, seed in zip(batch_messages, seeds)
        ]
    if print_progress:
        responses = await tqdm_asyncio.gather(*tasks, desc="Processing Prompts")
    else:
        responses = await asyncio.gather(
            *tasks, return_exceptions=True, desc="Processing Prompts"
        )

    final_results = []
    for res in responses:
        if isinstance(res, Exception):
            print(f"A request failed permanently after all retries: {res}")
            final_results.append(f"Error: {res}")
        else:
            final_results.append(res.choices[0].message.content)

    return final_results


def batch_decoding(
    model: Union[LLM, AsyncOpenAI],
    prompts: List[str] = ["Hi there! What is your name?"],
    stop_tokens: List[str] = ["\nA:"],
    structured_output_options: Optional[
        Union[ResponseGenerationMethod, List[ResponseGenerationMethod]]
    ] = None,
    seed: int = 42,
    client_model_name: Optional[str] = None,
    api_concurrency: int = 10,
    print_conversation: bool = False,
    print_progress: bool = True,
    **generation_kwargs: Any,
):
    """
    Generate responses for a batch of prompts.

    Handles both vLLM and OpenAI API generation with support for:
    - Structured output (JSON or choice format)
    - Conversation printing
    - Progress tracking
    - Concurrent API requests

    Args:
        model: vLLM model or AsyncOpenAI client
        system_messages: System prompts for each conversation
        prompts: User prompts to generate responses for
        structured_output_options: Configuration for structured output
        seed: Random seed for reproducibility
        client_model_name: Model name when using OpenAI API
        api_concurrency: Max concurrent API requests
        print_conversation: If True, prints conversations
        print_progress: If True, shows progress bar
        **generation_kwargs: Additional generation parameters

    Returns:
        List[str]: Generated responses
    """
    random.seed(seed)

    batch_size: int = len(prompts)

    seeds = _generate_seeds(seed, batch_size=batch_size)

    if isinstance(model, LLM):
        sampling_params_list = _create_sampling_params(
            batch_size=batch_size,
            seeds=seeds,
            structured_output_options=structured_output_options,
            stop_tokens=stop_tokens,
            **generation_kwargs,
        )
        outputs: List[RequestOutput] = model.generate(
            prompts,
            sampling_params=sampling_params_list,
            use_tqdm=print_progress,
        )
        result = [output.outputs[0].text for output in outputs]

    else:
        result = _run_async_in_thread(
            client=model,
            client_model_name=client_model_name,
            batch_messages=prompts,
            seeds=seeds,
            concurrency_limit=api_concurrency,
            structured_output_options=structured_output_options,
            **generation_kwargs,
        )

    # TODO add argurment to specify how many conversations should be printed (base argument should be reasonable)
    if print_conversation:
        conversation_print = "Conversation:"
        for prompt, answer in zip(prompts, result):
            round_print = f"{conversation_print}\nUser Message:\n{prompt}\nGenerated Message\n{answer}"
            print(round_print, flush=True)
            break

    return result



from typing import Final, List, Dict

from enum import Enum

# Survey Item
SYSTEM_PROMPT_FIELD: Final[str] = "system_prompt"
INTERVIEW_INSTRUCTION_FIELD: Final[str] = "interview_instruction"
INTERVIEW_ITEM_ID: Final[str] = "interview_item_id"
INTERVIEW_ITEM: Final[str] = "interview_item"
INTERVIEW_NAME: Final[str] = "interview_name"

# Question
QUESTION_STEM: Final[str] = "question_stem"
QUESTION_CONTENT: Final[str] = "question_content"
QUESTION: Final[str] = "question"

# Answer
ANSWER_CODE: Final[str] = "answer_code"
ANSWER_TEXT: Final[str] = "answer_text"
ANSWER_OPTION: Final[str] = "answer_option"

# LLM Response
LLM_RESPONSE: Final[str] = "llm_response"
PARSED_RESPONSE: Final[str] = "parsed_response"

# Structured Output constraints
OPTIONS_ADJUST: List[str] = ["OPTIONS_ADJUST"]


class InterviewType(Enum):
    QUESTION: str = "interview_type_question"
    CONTEXT: str = "interview_type_context"
    ONE_PROMPT: str = "interview_type_one_prompt"


DEFAULT_SYSTEM_PROMPT: Final[str] = (
    "You will be given questions and possible answer options for each. Please reason about each question before answering."
)
DEFAULT_TASK_INSTRUCTION: Final[str] = ""
DEFAULT_JSON_STRUCTURE: Final[List[str]] = ["reasoning", "answer"]
DEFAULT_CONSTRAINTS: Final[Dict[str, List[str]]] = {"answer": OPTIONS_ADJUST}
DEFAULT_INTERVIEW_ID: Final[str] = "Interview"



"""
Module for managing and conducting surveys using LLM models.

This module provides functions to conduct surveys in different ways:
- Question by question
- Whole survey in one prompt
- In-context learning

Usage example:
-------------
```python
from surveygen import LLMInterview, conduct_survey_question_by_question
from surveygen.parser.llm_answer_parser import raw_responses
from vllm import LLM

# Initialize model and interview
model = LLM(model="meta-llama/Meta-Llama-3-8B-Instruct")
interview = LLMInterview(interview_path="questions.csv")
interview.prepare_interview(question_stem="How do you feel towards {QUESTION_CONTENT_PLACEHOLDER}?")

# Conduct survey
results = conduct_survey_question_by_question(
    model=model,
    interviews=interview,
    print_progress=True
)

# Access results
for result in results:
    raw_responses = raw_responses(survey_answers)
```
"""

from typing import (
    List,
    Dict,
    Optional,
    Union,
    Any,
    Literal,
)
from string import ascii_lowercase, ascii_uppercase

from .utilities.survey_objects import (
    AnswerOptions,
    QuestionLLMResponseTuple,
    AnswerTexts,
)
from .utilities import prompt_templates
from .utilities import constants
from .utilities import utils

from .parser.llm_answer_parser import raw_responses

from .inference.survey_inference import batch_generation, batch_turn_by_turn_generation
from .inference.response_generation import (
    ResponseGenerationMethod,
    JSONResponseGenerationMethod,
)

from .llm_interview import LLMInterview, InterviewType

from .utilities.survey_objects import AnswerOptions, InterviewResult

from vllm import LLM

from openai import AsyncOpenAI

from pathlib import Path
import os

import pandas as pd

import random

from tqdm.auto import tqdm


class SurveyOptionGenerator:
    """
    This class offers robust creation of options. Can do various prompt pertubations.
    When used in Conjunction with response generation methods the tokens for the output of the model can be restricted.
    """

    LIKERT_5: List[str] = [
        "disagree strongly",
        "disagree a little",
        "neither agree nor disagree",
        "agree a little",
        "agree strongly",
    ]
    LIKERT_NO_MIDDLE: List[str] = [
        "disagree strongly",
        "disagree a little",
        "agree a little",
        "agree strongly",
    ]

    LIKERT_IMPORTANCE_FROM_TO: List[str] = ["Not at all important", "Very Important"]
    LIKERT_JUSTIFIABLE_FROM_TO: List[str] = ["Never justifiable", "Always justifiable"]
    _IDX_TYPES = Literal["char_lower", "char_upper", "integer", "no_index"]

    @staticmethod
    def generate_likert_options(
        n: int,
        answer_texts: Optional[List[str]],
        only_from_to_scale: bool = False,
        random_order: bool = False,
        reversed_order: bool = False,
        even_order: bool = False,
        start_idx: int = 1,
        list_prompt_template: str = prompt_templates.LIST_OPTIONS_DEFAULT,
        scale_prompt_template: str = prompt_templates.SCALE_OPTIONS_DEFAULT,
        index_answer_separator: str = ": ",
        options_separator: str = ", ",
        idx_type: _IDX_TYPES = "integer",
        response_generation_method: Optional[ResponseGenerationMethod] = None,
    ) -> AnswerOptions:
        """Generates a set of options and a prompt for a Likert-style scale.

        This function creates a numeric or alphabetic scale of a specified size (n),
        optionally attaching textual labels to the scale. It provides
        extensive control over ordering, formatting, and the final prompt string.

        Args:
            n (int): The number of options to generate (e.g., 5 for a 5-point scale).
            answer_texts (Optional[List[str]]): A list of text labels for each option.
                Its length must equal `n` if provided.
            only_from_to_scale (bool, optional): If True, the prompt will only show the
                min and max of the scale (e.g., "1 to 5"). Defaults to False.
            random_order (bool, optional): If True, the options are randomized. Defaults to False.
            reversed_order (bool, optional): If True, the options are in reversed input order.
                Defaults to False.
            even_order (bool, optional): If True, options the center option will be removed.
                E.g., for n=5: 1, 2, 4, 5
            start_idx (int, optional): The starting index for the scale (usually 0 or 1).
                Defaults to 1.
            list_prompt_template (str, optional): The template for prompts that list all options.
            scale_prompt_template (str, optional): The template for prompts that only show the range.
            index_answer_separator (str, optional): The string used to separate an index from its
                text label (e.g., "1: Strongly Agree"). Defaults to ": ".
            options_separator (str, optional): The string used to separate options when listed
                in the prompt. Defaults to ", ".
            idx_type (_IDX_TYPES, optional): The type of index to use: "integer", "upper" (A, B, C),
                or "lower" (a, b, c). Defaults to "integer".
            response_generation_method (Optional[ResponseGenerationMethod], optional): An object
                controlling how the final response object is generated. Defaults to None.

        Raises:
            ValueError: If `answer_texts` is provided and its length does not match `n`.

        Returns:
            AnswerOptions: An object containing the generated list of option strings and the
            final formatted prompt ready for display.

        Example:
            .. code-block:: python

                # Generate a classic 5-point "Strongly Disagree" to "Strongly Agree" scale
                labels = [
                    "Strongly Disagree", "Disagree", "Neutral", "Agree", "Strongly Agree"
                ]
                options = SurveyOptionGenerator.generate_likert_options(n=5, answer_texts=labels)
        """

        if only_from_to_scale:
            if len(answer_texts) != 2:
                raise ValueError(
                    f"From-To scales require exactly 2 descriptions, but answer_texts was set to '{answer_texts}'."
                )
            if idx_type != "integer":
                raise ValueError(
                    f"From-To scales require an integer scale index, but idx_type was set to '{idx_type}'."
                )
        else:
            if answer_texts:
                if len(answer_texts) != n:
                    raise ValueError(
                        f"answer_texts and n need to be the same length, but answer_texts has length {len(answer_texts)} and n was given as {n}."
                    )
        if even_order:
            if n % 2 != 0:
                raise ValueError(
                    "If you want to turn a scale even, it should be odd before."
                )
            middle_index = n // 2
            answer_texts = (
                answer_texts[:middle_index] + answer_texts[middle_index + 1 :]
            )
            n = n - 1
        if random_order:
            if len(answer_texts) < 2:
                raise ValueError(
                    "There must be at least two answer options to reorder randomly."
                )
            random.shuffle(
                answer_texts
            )  # no assignment needed because shuffles already inplace
        if reversed_order:
            if len(answer_texts) < 2:
                raise ValueError(
                    "There must be at least two answer options to reorder in reverse."
                )
            answer_options = answer_options[::-1]

        answer_option_indices = []
        if idx_type == "no_index":
            # no index, just the answer options directly
            answer_option_indices = None
        elif idx_type == "integer":
            for i in range(n):
                answer_code = i + start_idx
                answer_option_indices.append(str(answer_code))
        else:
            # TODO @Jens add these to constants.py
            if idx_type == "char_lower":
                for i in range(n):
                    answer_option_indices.append(ascii_lowercase[(i + start_idx) % 26])
            elif idx_type == "char_upper":
                for i in range(n):
                    answer_option_indices.append(ascii_uppercase[(i + start_idx) % 26])

        answer_texts_object = AnswerTexts(
            answer_texts=answer_texts,
            indices=answer_option_indices,
            index_answer_seperator=index_answer_separator,
            option_seperators=options_separator,
            only_scale=only_from_to_scale,
        )

        interview_option = AnswerOptions(
            answer_texts=answer_texts_object,
            from_to_scale=only_from_to_scale,
            list_prompt_template=list_prompt_template,
            scale_prompt_template=scale_prompt_template,
            response_generation_method=response_generation_method,
        )

        return interview_option

    # #TODO: It seems to me like this method and the one above could be merged? (Georg)
    # @staticmethod
    # def generate_generic_options(
    #     answer_texts: Dict,
    #     only_from_to_scale: bool = False,
    #     random_order: bool = False,
    #     reversed_order: bool = False,
    #     even_order: bool = False,
    #     idx_type: Optional[_IDX_TYPES] = None, # uses the answer_texts.keys() as an index by default
    #     list_prompt_template: Optional[str] = prompt_templates.LIST_OPTIONS_DEFAULT,
    #     scale_prompt_template: Optional[str] = prompt_templates.SCALE_OPTIONS_DEFAULT,
    #     options_separator: str = ", ",
    # ):

    #     n = len(answer_texts.values())
    #     answer_codes = answer_texts.keys()
    #     answer_texts = answer_texts.values()
    #     # answer_options = descriptions

    #     if idx_type == 'char_lower':
    #         if all(isinstance(item, int) for item in answer_codes):
    #             new_codes = []
    #             for i in answer_codes:
    #                 code = ascii_lowercase[i - 1]
    #                 new_codes.append(code)
    #             answer_codes = new_codes
    #         else:
    #             answer_codes = [s.lower() for s in answer_codes]
    #     elif idx_type == 'char_upper':
    #         if all(isinstance(item, int) for item in answer_codes):
    #             new_codes = []
    #             for i in answer_codes:
    #                 code = ascii_uppercase[i - 1]
    #                 new_codes.append(code)
    #             answer_codes = new_codes
    #         else:
    #             answer_codes = [s.upper() for s in answer_codes]
    #     elif idx_type == 'integer':
    #         answer_codes = range(1, len(answer_codes) + 1)

    #     answer_options = dict(zip(answer_codes, answer_texts))

    #     if only_from_to_scale:
    #         assert all(
    #             isinstance(item, int) for item in answer_codes
    #         ), "To use from-to scale you must have integer answer codes."

    #     if random_order:
    #         assert (
    #             n >= 2
    #         ), "There must be at least two answer options to reorder randomly."
    #         temp = list(answer_texts)
    #         random.shuffle(temp)
    #         # reassigning to keys
    #         answer_options = dict(zip(answer_codes, temp))
    #     if reversed_order:
    #         assert (
    #             n >= 2
    #         ), "There must be at least two answer options to reverse options."
    #         reversed_values = list(answer_texts)[::-1]
    #         answer_options = dict(zip(answer_codes, reversed_values))
    #     if even_order:
    #         assert n % 2 != 0, "There must be a odd number of options!"
    #         middle_index = n // 2
    #         # Get the key of the item to be removed
    #         key_to_remove = list(answer_codes)[middle_index]
    #         # Create a new dictionary, excluding the item with key_to_remove
    #         # This uses a dictionary comprehension.
    #         answer_options = {
    #             key: value
    #             for key, value in answer_options.items()
    #             if key != key_to_remove
    #         }
    #         if all(isinstance(element, int) for element in list(answer_codes)):
    #             first_part = list(answer_codes)[:middle_index]
    #             last_part = list(answer_codes)[middle_index + 1 :]
    #             last_part = [x - 1 for x in last_part]
    #             answer_options = dict(
    #                 zip(first_part + last_part, answer_options.values())
    #             )
    #         elif set(list(answer_codes)).issubset(list(ascii_lowercase)):
    #             first_part = list(answer_codes)[:middle_index]
    #             # print("First part:", first_part)
    #             last_part = list(answer_codes)[middle_index + 1 :]
    #             # print("Last part:", last_part)
    #             last_parts = []
    #             for i in range(middle_index + 1, len(list(answer_codes))):
    #                 part = list(ascii_lowercase)[i - 1]
    #                 last_parts.append(part)
    #                 # print("Last parts:", last_parts)
    #             answer_options = dict(
    #                 zip(first_part + last_parts, answer_options.values())
    #             )
    #         elif set(list(answer_codes)).issubset(list(ascii_uppercase)):
    #             first_part = list(answer_codes)[:middle_index]
    #             # print("First part:", first_part)
    #             last_part = list(answer_codes)[middle_index + 1 :]
    #             # print("Last part:", last_part)
    #             last_parts = []
    #             for i in range(middle_index + 1, len(list(answer_codes))):
    #                 part = list(ascii_uppercase)[i - 1]
    #                 last_parts.append(part)
    #                 # print("Last parts:", last_parts)
    #             answer_options = dict(
    #                 zip(first_part + last_parts, answer_options.values())
    #             )

    #     if idx_type == 'no_index':
    #         answer_option_strings = list(answer_options.values())
    #         answer_option_index = None
    #     else:
    #         answer_option_strings = [f"{key}: {val}" for key, val in answer_options.items()]
    #         answer_option_index = list(answer_options.keys())
    #     #print(answer_options)

    #     interview_option = AnswerOptions(
    #         answer_text = answer_option_strings,
    #         index = answer_option_index,
    #         from_to_scale = only_from_to_scale,
    #         list_prompt_template = list_prompt_template,
    #         scale_prompt_template = scale_prompt_template,
    #         options_seperator = options_separator,
    #     )

    #     return interview_option


def conduct_survey_single_item(
    model: Union[LLM, AsyncOpenAI],
    interviews: Union[LLMInterview, List[LLMInterview]],
    client_model_name: Optional[str] = None,
    api_concurrency: int = 10,
    print_conversation: bool = False,
    print_progress: bool = True,
    n_save_step: Optional[int] = None,
    intermediate_save_file: Optional[str] = None,
    seed: int = 42,
    chat_template: Optional[str] = None,
    chat_template_kwargs: Dict[str, Any] = {},
    **generation_kwargs: Any,
) -> List[InterviewResult]:
    """
    Conducts a survey by asking questions one at a time.

    Args:
        model: LLM instance or AsyncOpenAI client.
        interviews: Single interview or list of interviews to conduct as a survey.
        answer_production_method: Options for structured output format.
        client_model_name: Name of model when using OpenAI client.
        api_concurrency: Number of concurrent API requests.
        print_conversation: If True, prints all conversations.
        print_progress: If True, shows progress bar.
        n_save_step: Save intermediate results every n steps.
        intermediate_save_file: Path to save intermediate results.
        seed: Random seed for reproducibility.
        chat_template: Optionally pass a specific chat template
        chat_template_kwargs: Arguments to pass to the chat template, e.g., to disable reasoning
        **generation_kwargs: Additional generation parameters that will be given to vllm.chat() or  client.chat.completions.create().

    Returns:
        List[InterviewResult]: Results for each interview.
    """

    _intermediate_save_path_check(n_save_step, intermediate_save_file)

    if isinstance(interviews, LLMInterview):
        interviews = [interviews]

    max_survey_length: int = max(len(interview._questions) for interview in interviews)
    question_llm_response_pairs: List[Dict[int, QuestionLLMResponseTuple]] = []

    for i in range(len(interviews)):
        # inference_option = interviews[i]._generate_inference_options()
        # inference_options.append(inference_opti
        question_llm_response_pairs.append({})

    survey_results: List[InterviewResult] = []

    for i in (
        tqdm(range(max_survey_length), desc="Processing interviews")
        if print_progress
        else range(max_survey_length)
    ):
        current_batch: List[LLMInterview] = [
            interview for interview in interviews if len(interview._questions) > i
        ]

        system_messages, prompts = zip(
            *[
                interview.get_prompt_for_interview_type(InterviewType.QUESTION, i)
                for interview in current_batch
            ]
        )

        questions = [
            interview.generate_question_prompt(
                interview_question=interview._questions[i]
            )
            for interview in current_batch
        ]
        response_generation_methods = [
            (
                interview._questions[i].answer_options.response_generation_method
                if interview._questions[i].answer_options
                else None
            )
            for interview in current_batch
        ]

        output, logprobs, reasoning_output = batch_generation(
            model=model,
            system_messages=system_messages,
            prompts=prompts,
            response_generation_method=response_generation_methods,
            client_model_name=client_model_name,
            api_concurrency=api_concurrency,
            print_conversation=print_conversation,
            print_progress=print_progress,
            seed=seed,
            chat_template=chat_template,
            chat_template_kwargs=chat_template_kwargs,
            **generation_kwargs,
        )

        # avoid errors when zipping
        if logprobs is None:
            logprobs = [None] * len(current_batch)

        for survey_id, question, answer, logprob_answer, reasoning, item in zip(
            range(len(current_batch)),
            questions,
            output,
            logprobs,
            reasoning_output,
            current_batch,
        ):
            question_llm_response_pairs[survey_id].update(
                {
                    item._questions[i].item_id: QuestionLLMResponseTuple(
                        question, answer, logprob_answer, reasoning
                    )
                }
            )

        # TODO: check that this works with logprobs
        _intermediate_saves(
            interviews,
            n_save_step,
            intermediate_save_file,
            question_llm_response_pairs,
            i,
        )

    for i, survey in enumerate(interviews):
        survey_results.append(InterviewResult(survey, question_llm_response_pairs[i]))

    return survey_results


def _intermediate_saves(
    interviews: List[LLMInterview],
    n_save_step: int,
    intermediate_save_file: str,
    question_llm_response_pairs: QuestionLLMResponseTuple,
    i: int,
):
    """
    Internal helper to save intermediate survey results.

    Args:
        interviews: List of interviews being conducted.
        n_save_step: Save frequency in steps.
        intermediate_save_file: Path to save file.
        question_llm_response_pairs: Current responses.
        i: Current step number.
    """
    if n_save_step:
        if i % n_save_step == 0:
            intermediate_survey_results: List[InterviewResult] = []
            for j, interview in enumerate(interviews):
                intermediate_survey_results.append(
                    InterviewResult(interview, question_llm_response_pairs[j])
                )
            parsed_results = raw_responses(intermediate_survey_results)
            utils.create_one_dataframe(parsed_results).to_csv(intermediate_save_file)


def _intermediate_save_path_check(n_save_step: int, intermediate_save_path: str):
    """
    Internal helper to validate intermediate save path.

    Args:
        n_save_step: Save frequency in steps.
        intermediate_save_path: Path to check.
    """
    if n_save_step:
        if not isinstance(n_save_step, int) or n_save_step <= 0:
            raise ValueError("`n_save_step` must be a positive integer.")

        if not intermediate_save_path:
            raise ValueError(
                "`intermediate_save_file` must be provided if saving is enabled."
            )

        if not intermediate_save_path.endswith(".csv"):
            raise ValueError("`intermediate_save_file` should be a .csv file.")

        # Ensure it's a directory that exists or can be created
        parent_dir = Path(intermediate_save_path).parent
        if not parent_dir.exists():
            try:
                parent_dir.mkdir(parents=True, exist_ok=True)
            except Exception as e:
                raise ValueError(
                    f"Invalid intermediate save path: {intermediate_save_path}. Error: {e}"
                )

        # Optional: Check it's writable
        if not os.access(parent_dir, os.W_OK):
            raise ValueError(f"Save path '{intermediate_save_path}' is not writable.")


def conduct_survey_battery(
    model: Union[LLM, AsyncOpenAI],
    interviews: Union[LLMInterview, List[LLMInterview]],
    client_model_name: Optional[str] = None,
    api_concurrency: int = 10,
    n_save_step: Optional[int] = None,
    intermediate_save_file: Optional[str] = None,
    print_conversation: bool = False,
    print_progress: bool = True,
    seed: int = 42,
    chat_template: Optional[str] = None,
    chat_template_kwargs: Dict[str, Any] = {},
    item_separator: str = "\n",
    **generation_kwargs: Any,
) -> List[InterviewResult]:
    """
    Conducts the entire survey in one single LLM prompt.

    Args:
        model: LLM instance or AsyncOpenAI client.
        interviews: Single interview or list of interviews to conduct.
        answer_production_method: Options for structured output format.
        client_model_name: Name of model when using OpenAI client.
        api_concurrency: Number of concurrent API requests.
        n_save_step: Save intermediate results every n steps.
        intermediate_save_file: Path to save intermediate results.
        print_conversation: If True, prints the conversation.
        print_progress: If True, shows progress bar.
        seed: Random seed for reproducibility.
        chat_template: Optionally pass a specific chat template
        chat_template_kwargs: Arguments to pass to the chat template, e.g., to disable reasoning
        **generation_kwargs: Additional generation parameters that will be given to vllm.chat() or  client.chat.completions.create().

    Returns:
        List[InterviewResult]: Results for each interview.
    """
    _intermediate_save_path_check(n_save_step, intermediate_save_file)

    if isinstance(interviews, LLMInterview):
        interviews = [interviews]
    # inference_options: List[InferenceOptions] = []

    # We always conduct the survey in one prompt
    max_survey_length: int = 1

    question_llm_response_pairs: List[Dict[int, QuestionLLMResponseTuple]] = []

    # if print_progress:
    #     print("Constructing prompts")
    for i in range(len(interviews)):
        question_llm_response_pairs.append({})

    survey_results: List[InterviewResult] = []

    for i in (
        tqdm(range(max_survey_length), desc="Processing interviews")
        if print_progress
        else range(max_survey_length)
    ):
        current_batch = [
            interview for interview in interviews if len(interview._questions) > i
        ]

        system_messages, prompts = zip(
            *[
                interview.get_prompt_for_interview_type(InterviewType.ONE_PROMPT, i)
                for interview in current_batch
            ]
        )
        # questions = [interview.generate_question_prompt(interview_question=interview._questions[i]) for interview in current_batch]
        response_generation_methods: List[ResponseGenerationMethod] = []
        for interview in current_batch:
            if interview._questions[i].answer_options:
                response_generation_method = interview._questions[
                    i
                ].answer_options.response_generation_method
                if isinstance(response_generation_method, JSONResponseGenerationMethod):
                    response_generation_method = response_generation_method.create_new_rgm_with_multiple_questions(
                        questions=interview._questions
                    )
                response_generation_methods.append(response_generation_method)

        output, logprobs, reasoning_output = batch_generation(
            model=model,
            system_messages=system_messages,
            prompts=prompts,
            response_generation_method=response_generation_methods,
            client_model_name=client_model_name,
            api_concurrency=api_concurrency,
            print_conversation=print_conversation,
            print_progress=print_progress,
            seed=seed,
            chat_template=chat_template,
            chat_template_kwargs=chat_template_kwargs,
            **generation_kwargs,
        )

        # avoid errors when zipping
        if logprobs is None:
            logprobs = [None] * len(current_batch)

        for survey_id, prompt, answer, logprob_answer, reasoning in zip(
            range(len(current_batch)), prompts, output, logprobs, reasoning_output
        ):
            question_llm_response_pairs[survey_id].update(
                {
                    -1: QuestionLLMResponseTuple(
                        prompt, answer, logprob_answer, reasoning
                    )
                }
            )

        _intermediate_saves(
            interviews,
            n_save_step,
            intermediate_save_file,
            question_llm_response_pairs,
            i,
        )

    for i, survey in enumerate(interviews):
        survey_results.append(InterviewResult(survey, question_llm_response_pairs[i]))

    return survey_results


def conduct_survey_sequential(
    model: Union[LLM, AsyncOpenAI],
    interviews: Union[LLMInterview, List[LLMInterview]],
    client_model_name: Optional[str] = None,
    api_concurrency: int = 10,
    print_conversation: bool = False,
    print_progress: bool = True,
    n_save_step: Optional[int] = None,
    intermediate_save_file: Optional[str] = None,
    seed: int = 42,
    **generation_kwargs: Any,
) -> List[InterviewResult]:
    """
    Conducts surveys using in-context learning approach.

    Args:
        model: LLM instance or AsyncOpenAI client.
        interviews: Single interview or list of interviews to conduct.
        answer_production_method: Options for structured output format.
        client_model_name: Name of model when using OpenAI client.
        api_concurrency: Number of concurrent API requests.
        print_conversation: If True, prints the conversation.
        print_progress: If True, shows progress bar.
        n_save_step: Save intermediate results every n steps.
        intermediate_save_file: Path to save intermediate results.
        seed: Random seed for reproducibility.
        **generation_kwargs: Additional generation parameters that will be given to vllm.chat() or  client.chat.completions.create().

    Returns:
        List[InterviewResult]: Results for each interview.
    """
    _intermediate_save_path_check(n_save_step, intermediate_save_file)
    if isinstance(interviews, LLMInterview):
        interviews = [interviews]

    max_survey_length: int = max(len(interview._questions) for interview in interviews)

    question_llm_response: List[Dict[int, QuestionLLMResponseTuple]] = []

    for i in range(len(interviews)):
        question_llm_response.append({})

    survey_results: List[InterviewResult] = []

    all_prompts: List[List[str]] = []
    assistant_messages: List[List[str]] = []

    for i in range(len(interviews)):
        assistant_messages.append([])
        all_prompts.append([])

    for i in (
        tqdm(range(max_survey_length), desc="Processing interviews")
        if print_progress
        else range(max_survey_length)
    ):
        current_batch = [
            interview for interview in interviews if len(interview._questions) > i
        ]

        first_question: bool = i == 0

        if first_question:
            system_messages, prompts = zip(
                *[
                    interview.get_prompt_for_interview_type(InterviewType.CONTEXT, i)
                    for interview in current_batch
                ]
            )
            questions = [
                interview.generate_question_prompt(
                    interview_question=interview._questions[i]
                )
                for interview in current_batch
            ]
        else:
            system_messages, _ = zip(
                *[
                    interview.get_prompt_for_interview_type(InterviewType.CONTEXT, i)
                    for interview in current_batch
                ]
            )
            prompts = [
                interview.generate_question_prompt(
                    interview_question=interview._questions[i]
                )
                for interview in current_batch
            ]
            questions = prompts

        response_generation_methods = [
            (
                interview._questions[i].answer_options.response_generation_method
                if interview._questions[i].answer_options
                else None
            )
            for interview in current_batch
        ]

        for c in range(len(current_batch)):
            all_prompts[c].append(prompts[c])

        current_assistant_messages: List[str] = []

        missing_indeces = []

        for index, surv in enumerate(current_batch):
            prefilled_answer = surv._questions[i].prefilled_response
            if prefilled_answer:
                current_assistant_messages.append(prefilled_answer)
                missing_indeces.append(index)

        current_batch = [
            item for a, item in enumerate(current_batch) if a not in missing_indeces
        ]

        if len(current_batch) == 0:
            for c in range(len(current_batch)):
                assistant_messages[c].append(current_assistant_messages[c])
            for (
                survey_id,
                question,
                llm_response,
                logprob_answer,
                reasoning,
                item,
            ) in zip(
                range(len(current_batch)),
                questions,
                current_assistant_messages,
                logprobs,
                reasoning_output,
                current_batch,
            ):
                question_llm_response[survey_id].update(
                    {
                        item._questions[i].item_id: QuestionLLMResponseTuple(
                            question, llm_response, logprob_answer, reasoning
                        )
                    }
                )
            continue
            # TODO: add support for automatic system prompt for other answer production methods

        output, logprobs, reasoning_output = batch_turn_by_turn_generation(
            model=model,
            system_messages=system_messages,
            prompts=all_prompts,
            assistant_messages=assistant_messages,
            response_generation_method=response_generation_methods,
            client_model_name=client_model_name,
            api_concurrency=api_concurrency,
            print_conversation=print_conversation,
            print_progress=print_progress,
            seed=seed,
            **generation_kwargs,
        )

        # avoid errors when zipping
        if logprobs is None or len(logprobs) == 0:
            logprobs = [None] * len(current_batch)

        for num, index in enumerate(missing_indeces):
            output.insert(index, current_assistant_messages[num])
        for survey_id, question, llm_response, logprob_answer, reasoning, item in zip(
            range(len(current_batch)),
            questions,
            output,
            logprobs,
            reasoning_output,
            current_batch,
        ):
            question_llm_response[survey_id].update(
                {
                    item._questions[i].item_id: QuestionLLMResponseTuple(
                        question, llm_response, logprob_answer, reasoning
                    )
                }
            )

        for o in range(len(output)):
            assistant_messages[o].append(output[o])
        # assistant_messages.append(output)

        _intermediate_saves(
            interviews, n_save_step, intermediate_save_file, question_llm_response, i
        )

    for i, survey in enumerate(interviews):
        survey_results.append(InterviewResult(survey, question_llm_response[i]))

    return survey_results


class SurveyCreator:
    @classmethod
    def from_path(
        self, survey_path: str, questionnaire_path: str
    ) -> List[LLMInterview]:
        """
        Generates LLMInterview objects from a CSV file path.

        Args:
            survey_path: The path to the CSV file.

        Returns:
            A list of LLMInterview objects.
        """
        df = pd.read_csv(survey_path)
        df_questionnaire = pd.read_csv(questionnaire_path)
        return self._from_dataframe(df, df_questionnaire)

    @classmethod
    def from_dataframe(
        self, survey_dataframe: pd.DataFrame, questionnaire_dataframe: pd.DataFrame
    ) -> List[LLMInterview]:
        """
        Generates LLMInterview objects from a pandas DataFrame.

        Args:
            survey_dataframe: A DataFrame containing the survey data.

        Returns:
            A list of LLMInterview objects.
        """
        return self._from_dataframe(survey_dataframe, questionnaire_dataframe)

    @classmethod
    def _create_interview(self, row: pd.Series, df_questionnaire) -> LLMInterview:
        """
        Internal helper method to process the DataFrame.
        """
        return LLMInterview(
            interview_source=df_questionnaire,
            interview_name=row[constants.INTERVIEW_NAME],
            system_prompt=row[constants.SYSTEM_PROMPT_FIELD],
            prompt=row[constants.INTERVIEW_INSTRUCTION_FIELD],
        )

    @classmethod
    def _from_dataframe(
        self, df: pd.DataFrame, df_questionnaire: pd.DataFrame
    ) -> List[LLMInterview]:
        """
        Internal helper method to process the DataFrame.
        """
        interviews = df.apply(
            lambda row: self._create_interview(row, df_questionnaire), axis=1
        )
        return interviews.to_list()


if __name__ == "__main__":
    pass



from setuptools import setup, find_packages

setup(
    name="qstn",
    version="0.1.0",
    packages=find_packages(),
    install_requires=open("requirements.txt").read().splitlines(),
    author="Maximilian Kreutner",
    author_email="maximilian.kreutner@uni-mannheim.de",
    description="The QSTN Package",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.11',
)


from typing import List, Dict, Optional, Union, overload, Self, Tuple

from dataclasses import replace

from .utilities.survey_objects import AnswerOptions, InterviewItem

from .utilities import constants, placeholder
from .utilities.constants import InterviewType

from .utilities.utils import safe_format_with_regex

import pandas as pd

import random

import copy


from transformers import AutoTokenizer


class LLMInterview:
    """
    Main class for setting up and managing an LLM-based interview or survey.

    This class handles loading questions, preparing prompts, managing answer options,
    and generating prompt structures for different interview types.
    """

    DEFAULT_INTERVIEW_ID: str = "Interview"

    DEFAULT_SYSTEM_PROMPT: str = (
        "You will be given questions and possible answer options for each. Please reason about each question before answering."
    )
    DEFAULT_TASK_INSTRUCTION: str = ""

    DEFAULT_JSON_STRUCTURE: List[str] = ["reasoning", "answer"]

    DEFAULT_PROMPT_STRUCTURE: str = (
        f"{placeholder.PROMPT_QUESTIONS}\n{placeholder.PROMPT_OPTIONS}"
    )

    def __init__(
        self,
        interview_source=Union[str, pd.DataFrame],
        interview_name: str = DEFAULT_INTERVIEW_ID,
        system_prompt: str = DEFAULT_SYSTEM_PROMPT,
        prompt: str = DEFAULT_PROMPT_STRUCTURE,
        verbose=False,
        seed: int = 42,
    ):
        """
        Initialize an LLMInterview instance. Either a path to a csv file or a pandas dataframe has to be provided

        Args:
            interview_path (str): Path to the CSV file containing the interview structure and questions.
            interview_dataframe (pd.Dataframe): A pandas dataframe interview structure and questions.
            interview_name (str): Name/ID for the interview.
            system_prompt (str): System prompt to prepend to all questions.
            interview_instruction (str): Instructions that will be given to the model before asking the questions.
            verbose (bool): If True, enables verbose output.
            seed (int): Random seed for reproducibility.
        """
        random.seed(seed)

        if interview_source is None:
            raise ValueError("Either a path or a dataframe have to be provided")

        self.load_interview_format(interview_source=interview_source)

        self.verbose: bool = verbose

        self.interview_name: str = interview_name

        self.system_prompt: str = system_prompt
        self.prompt: str = prompt

        self._same_options = False

    def duplicate(self):
        """
        Create a deep copy of the current interview instance.

        Returns:
            LLMInterview: A deep copy of the current object.
        """
        return copy.deepcopy(self)

    def get_prompt_for_interview_type(
        self,
        interview_type: InterviewType = InterviewType.QUESTION,
        item_id: int = 0,
        item_separator: str = "\n",
    ) -> Tuple[str, str]:
        """
        Generate the full prompt for a given interview type.

        Args:
            interview_type (InterviewType): The type of interview prompt to generate.

        Returns:
            str: The constructed prompt for the interview type.
        """
        options = ""
        automatic_output_instructions = ""
        if (
            interview_type == InterviewType.QUESTION
            or interview_type == InterviewType.CONTEXT
        ):
            question = self.generate_question_prompt(self._questions[item_id])

            if self._questions[item_id].answer_options:
                options = self._questions[item_id].answer_options.create_options_str()

                rgm = self._questions[item_id].answer_options.response_generation_method
                if rgm is None:  # by default, no response generation method is required
                    automatic_output_instructions = ""
                else:
                    automatic_output_instructions: str = rgm.get_automatic_prompt()
            else:
                options = ""
                automatic_output_instructions = ""

            format_dict = {
                placeholder.PROMPT_QUESTIONS: question,
                placeholder.PROMPT_OPTIONS: options,
                placeholder.PROMPT_AUTOMATIC_OUTPUT_INSTRUCTIONS: automatic_output_instructions,
            }

        elif interview_type == InterviewType.ONE_PROMPT:
            all_questions: List[str] = []
            for question in self._questions:
                current_question_prompt = self.generate_question_prompt(question)

                if question.answer_options:
                    options = question.answer_options.create_options_str()
                else:
                    options = ""
                format_dict = {
                    placeholder.PROMPT_OPTIONS: options,
                }
                current_question_prompt = safe_format_with_regex(
                    current_question_prompt, format_dict
                )
                all_questions.append(current_question_prompt)

            all_questions_str = item_separator.join(all_questions)
            if self._questions[item_id].answer_options:
                options = self._questions[item_id].answer_options.create_options_str()
                rgm = self._questions[item_id].answer_options.response_generation_method

                if rgm is None:  # by default, no response generation method is required
                    automatic_output_instructions = ""
                else:
                    automatic_output_instructions: str = rgm.get_automatic_prompt(
                        questions=self._questions
                    )
            else:
                options = ""
                automatic_output_instructions = ""

            format_dict = {
                placeholder.PROMPT_QUESTIONS: all_questions_str,
                placeholder.PROMPT_OPTIONS: options,
                placeholder.PROMPT_AUTOMATIC_OUTPUT_INSTRUCTIONS: automatic_output_instructions,
            }

        system_prompt = safe_format_with_regex(self.system_prompt, format_dict)
        prompt = safe_format_with_regex(self.prompt, format_dict)

        return system_prompt, prompt

    def calculate_input_token_estimate(
        self, model_id: str, interview_type: InterviewType = InterviewType.QUESTION
    ) -> int:
        """
        Estimate the number of input tokens for the prompt, given a model and interview type.
        Remember that the model also has to have enough context length to fit its own response
        in case of CONTEXT and ONE_PROMPT type.

        Args:
            model_id (str): Huggingface model id.
            interview_type (InterviewType): Type of interview prompt.

        Returns:
            int: Estimated number of input tokens.
        """
        tokenizer = AutoTokenizer.from_pretrained(model_id)
        system_prompt, prompt = self.get_prompt_for_interview_type(
            interview_type=interview_type
        )
        system_tokens = tokenizer.encode(system_prompt)
        tokens = tokenizer.encode(prompt)
        total_tokens = len(system_tokens) + len(tokens)

        return (
            total_tokens
            if interview_type != InterviewType.CONTEXT
            else len(total_tokens) * 3
        )

    def get_survey_questions(self) -> str:
        """
        Get the list of loaded interview questions.

        Returns:
            List[InterviewItem]: The loaded questions.
        """
        return self._questions

    def load_interview_format(self, interview_source: Union[str, pd.DataFrame]) -> Self:
        """
        Load interview questions from a CSV file.

        The CSV should have columns: interview_item_id, question_content
        Optionally it can also have question_stem.

        Args:
            interview_source (str or pd.Dataframe): Path to a valid CSV file or pd.Dataframe.

        Returns:
            Self: The updated instance with loaded questions.
        """
        interview_questions: List[InterviewItem] = []

        if interview_source is None:
            raise ValueError("Either a path or a dataframe have to be provided")

        if type(interview_source) == pd.DataFrame:
            df = interview_source
        else:
            df = pd.read_csv(interview_source)

        for _, row in df.iterrows():
            interview_item_id = row[constants.INTERVIEW_ITEM_ID]
            # if constants.QUESTION in df.columns:
            #     question = row[constants.QUESTION]
            if constants.QUESTION_CONTENT in df.columns:
                interview_question_content = row[constants.QUESTION_CONTENT]
            else:
                interview_question_content = None

            if constants.QUESTION_STEM in df.columns:
                interview_question_stem = row[constants.QUESTION_STEM]
            else:
                interview_question_stem = None

            generated_interview_question = InterviewItem(
                item_id=interview_item_id,
                question_content=interview_question_content,
                question_stem=interview_question_stem,
            )
            interview_questions.append(generated_interview_question)

        self._questions = interview_questions
        return self

    # TODO Item order could be given by ids
    @overload
    def prepare_interview(
        self,
        question_stem: Optional[str] = None,
        answer_options: Optional[AnswerOptions] = None,
        prefilled_responses: Optional[Dict[int, str]] = None,
        randomized_item_order: bool = False,
    ) -> Self: ...

    @overload
    def prepare_interview(
        self,
        question_stem: Optional[List[str]] = None,
        answer_options: Optional[Dict[int, AnswerOptions]] = None,
        prefilled_responses: Optional[Dict[int, str]] = None,
        randomized_item_order: bool = False,
    ) -> Self: ...

    def prepare_interview(
        self,
        question_stem: Optional[Union[str, List[str]]] = None,
        answer_options: Optional[Union[AnswerOptions, Dict[int, AnswerOptions]]] = None,
        prefilled_responses: Optional[Dict[int, str]] = None,
        randomized_item_order: bool = False,
    ) -> Self:
        """
        Prepare the interview by assigning question stems, answer options, and prefilled responses.

        Args:
            question_stem (str or List[str], optional): Single or list of question stems.
            answer_options (AnswerOptions or Dict[int, AnswerOptions], optional): Answer options for all or per question.
            global_options (bool): If True, the answer options will be specified once at the end of the task instructions. Otherwise, they will be specified once per question.
            prefilled_responses (Dict[int, str], optional): If you provide prefilled responses, they will be used
            to fill the answers instead of prompting the LLM for that question.
            randomized_item_order (bool): If True, randomize the order of questions.

        Returns:
            Self: The updated instance with prepared questions.
        """
        interview_questions: List[InterviewItem] = self._questions

        prompt_list = isinstance(question_stem, list)
        if prompt_list:
            assert len(question_stem) == len(
                interview_questions
            ), "If a list of question stems is given, length of prompt and survey questions have to be the same"

        options_dict = False

        if isinstance(answer_options, AnswerOptions):
            self._same_options = True
            options_dict = False
        elif isinstance(answer_options, Dict):
            self._same_options = False
            options_dict = True

        updated_questions: List[InterviewItem] = []

        if not prefilled_responses:
            prefilled_responses = {}
            # for survey_question in survey_questions:
            # prefilled_answers[survey_question.question_id] = None

        if not prompt_list and not options_dict:
            updated_questions = []
            for i in range(len(interview_questions)):
                new_interview_question = replace(
                    interview_questions[i],
                    question_stem=(
                        question_stem
                        if question_stem
                        else interview_questions[i].question_stem
                    ),
                    answer_options=answer_options,
                    prefilled_response=prefilled_responses.get(
                        interview_questions[i].item_id
                    ),
                )
                updated_questions.append(new_interview_question)

        elif not prompt_list and options_dict:
            for i in range(len(interview_questions)):
                new_interview_question = replace(
                    interview_questions[i],
                    question_stem=(
                        question_stem
                        if question_stem
                        else interview_questions[i].question_stem
                    ),
                    answer_options=answer_options.get(interview_questions[i].item_id),
                    prefilled_response=prefilled_responses.get(
                        interview_questions[i].item_id
                    ),
                )
                updated_questions.append(new_interview_question)

        elif prompt_list and not options_dict:
            for i in range(len(interview_questions)):
                new_interview_question = replace(
                    interview_questions[i],
                    question_stem=(
                        question_stem[i]
                        if question_stem
                        else interview_questions[i].question_stem
                    ),
                    answer_options=answer_options,
                    prefilled_response=prefilled_responses.get(
                        interview_questions[i].item_id
                    ),
                )
                updated_questions.append(new_interview_question)
        elif prompt_list and options_dict:
            for i in range(len(interview_questions)):
                new_interview_question = replace(
                    interview_questions[i],
                    question_stem=(
                        question_stem[i]
                        if question_stem
                        else interview_questions[i].question_stem
                    ),
                    answer_options=answer_options.get(interview_questions[i].item_id),
                    prefilled_response=prefilled_responses.get(
                        interview_questions[i].item_id
                    ),
                )
                updated_questions.append(new_interview_question)

        if randomized_item_order:
            random.shuffle(updated_questions)

        self._questions = updated_questions
        return self

    def generate_question_prompt(self, interview_question: InterviewItem) -> str:
        """
        Generate the prompt string for a single interview question.

        Args:
            interview_question (InterviewItem): The question to prompt.

        Returns:
            str: The formatted prompt for the question.
        """

        if interview_question.question_stem:
            if placeholder.QUESTION_CONTENT in interview_question.question_stem:
                format_dict = {
                    placeholder.QUESTION_CONTENT: interview_question.question_content
                }
                question_prompt = safe_format_with_regex(
                    interview_question.question_stem, format_dict
                )
            else:
                question_prompt = f"""{interview_question.question_stem} {interview_question.question_content}"""
        else:
            question_prompt = f"""{interview_question.question_content}"""
        if interview_question.answer_options:
            _options_str = interview_question.answer_options.create_options_str()
            if _options_str is not None:
                safe_formatter = {placeholder.PROMPT_OPTIONS: _options_str}
                question_prompt = safe_format_with_regex(
                    question_prompt, safe_formatter
                )
        return question_prompt